/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/authenticate.cc
 *  
 *  Description:
 *    Implementation of the common authentication mechanism.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#include <string>
#include <vector>

#include <sys/types.h>
#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include <sys/wait.h>

#include "authenticate.h"
#include "io/io.h"
#include "session.h"
#include "parsers/storage/storage.h"
#include "util/convert.h"

using namespace ::std;
using namespace Binc;

namespace {

  bool enteredjail = false;

  void enterJail(void)
  {
    Session &session = Session::getInstance();
    IO &logger = IOFactory::getInstance().get(2);

    // drop all privileges that we can drop and enter chroot jail
    session.globalconfig.setSection("Security");
    const string &jailpath = session.globalconfig["jail path"];
    const string &jailuser = session.globalconfig["jail user"];
    const string &jailgroup = session.globalconfig["jail group"];
    struct group *gr = getgrnam(jailgroup.c_str());
    struct passwd *pw = getpwnam(jailuser.c_str());

    setgroups(0, NULL);

    if (chroot(jailpath.c_str()) != 0)
      logger << "Warning: unable to enter jail path "
	     << toImapString(jailpath) << endl;
    else
      chdir("/");

    if (gr == NULL)
      logger << "Warning: invalid jail group " 
	     << toImapString(jailuser) << endl;
    else {
      setgid(gr->gr_gid);

      if (pw == NULL)
	logger << "Warning: invalid jail user " 
	       << toImapString(jailuser) << endl;
      else
	setuid(pw->pw_uid);
    }

    umask(0);
  }
}

// 0 = ok
// 1 = internal error
// 2 = failed
//------------------------------------------------------------------------
int Binc::authenticate(Depot &depot, const string &username,
		       const string &password)
{
  Session &session = Session::getInstance();
  IO &com = IOFactory::getInstance().get(1);
  IO &logger = IOFactory::getInstance().get(2);

  // read auth penalty from global config
  session.globalconfig.setSection("Authentication");
  string authpenalty = session.globalconfig["auth penalty"];

  session.setUserID(username);

  // export the session data
  session.exportToEnv();
  
  // The information supplied on descriptor 3 is a login name
  // terminated by \0, a password terminated by \0, a timestamp
  // terminated by \0, and possibly more data. There are no other
  // restrictions on the form of the login name, password, and
  // timestamp.
  int authintercom[2];
  int intercomw[2];
  int intercomr[2];

  if (pipe(authintercom) == -1) return 1;
  if (pipe(intercomw) == -1) return 1;
  if (pipe(intercomr) == -1) return 1;

  string timestamp;
  time_t t;
  char *c;
  t = time(NULL);
  if ((c = ctime(&t)) != NULL)
    timestamp = c;
  else
    timestamp = "unknown timestamp";

  // execute authentication module
  int childspid, result;
  if ((childspid = fork()) == 0) {

    close(authintercom[1]);
    close(intercomr[0]);
    close(intercomw[1]);

    dup2(intercomr[1], 1);
    dup2(intercomw[0], 0);

    dup2(authintercom[0], 3);
    session.exportToEnv();

    if (session.unparsedArgs[0] != NULL) {
      execv(session.unparsedArgs[0], &session.unparsedArgs[0]);
      logger << "[auth module] invocation of " << session.unparsedArgs[0] 
	     << " failed: " << strerror(errno)
	     << endl;
      logger.flushContent();
      exit(111);
    }
    
    logger << "[auth module] Missing mandatory -- in arg list,"\
      " before authenticator." << endl;
    logger.flushContent();
    exit(111);
  }

  close(authintercom[0]);

  int error = 0;

  // write the userid
  if (write(authintercom[1],
	    username.c_str(),
	    username.length()) != (int) username.length()) 
    error = 1;

  // terminate with a \0
  if (error == 0 && write(authintercom[1], "", 1) != 1)
    error = 1;

  // write the password
  if (error == 0 && write(authintercom[1],
			  password.c_str(),
			  password.length()) != (int) password.length())
    error = 1;

  // terminate with a \0
  if (error == 0 && write(authintercom[1], "", 1) != 1)
    error = 1;

  // write the timestamp
  if (error == 0 && write(authintercom[1],
			  timestamp.c_str(),
			  timestamp.length()) != (int) timestamp.length()) 
    error = 1;

  // terminate with a \0
  if (error == 0 && write(authintercom[1], "", 1) != 1)
    error = 1;

  // close the write channel. this is necessary for the checkpassword
  // module to see an EOF.
  close(authintercom[1]);
  close(intercomr[1]);
  close(intercomw[0]);

  fd_set rmask;
  FD_ZERO(&rmask);
  FD_SET(fileno(stdin), &rmask);
  FD_SET(intercomr[0], &rmask);
  
  int maxfd = intercomr[0];

  session.globalconfig.setSection("Session");
  int idletimeout = atoi(session.globalconfig["idle timeout"]);
  if (idletimeout < (30 * 60)) idletimeout = 30 * 60;
  bool abort = false;

  com.disableInputLimit();

  try {

    bool eof = false;
    while (!eof) {
      fd_set rtmp = rmask;
      struct timeval timeout;
      timeout.tv_sec = idletimeout;
      timeout.tv_usec = 0;

      int n = select(maxfd + 1, &rtmp, NULL, NULL, &timeout);

      if (n < 0) {
	if (errno == EINTR) {
	  logger << "Warning: Got zero data from select: " << strerror(errno) << endl;
	  break;
	}

	logger << "Error: invalid exit from select, " << strerror(errno) << endl;
	break;
      }

      if (n == 0) {
	logger << "No activity from client - shutting down." << endl;
	break;
      }

      if (FD_ISSET(fileno(stdin), &rtmp)) {
	do {
	  int ret = com.readChar();
	  if (ret == 0) {
	    logger << "Warning: Client has disconnected" << endl;
	    eof = true;
	    break;
	  } else if (ret == -1) {
	    logger << "Error: When reading from client: "
		   << strerror(errno) << endl;
	    eof = true;
	    break;
	  } else if (ret == -2);
	  else {
	    session.addReadBytes(1);
	    char c = (char)ret;
	    int w = write(intercomw[1], &c, 1);
	  
	    if (w != 1) {
	      logger << "Error: When writing client data to child: "
		     << strerror(errno)
		     << endl;
	      eof = true;
	      break;
	    }
	  }
	} while (com.pending());
      }

      if (FD_ISSET(intercomr[0], &rtmp)) {
	char buf[1024];
	int ret = read(intercomr[0], buf, sizeof(buf));
	if (ret == 0) {
	  // Main server has shut down
	  eof = true;
	  break;
	} else if (ret == -1) {
	  logger << "Error: When reading from main server: " << strerror(errno)
		 << endl;
	  eof = true;
	  break;
	} else {
	  if (enteredjail == false) {
	    enterJail();
	    enteredjail = true;
	  }

	  session.addWriteBytes(ret);

	  com << string(buf, ret);
	  com.flushContent();
	}
      }
    }
  } catch (BincException &e) {
    logger << "Error, shutting down: " << e.getReason() << endl;
    abort = true;
  } catch (exception &e) {
    logger << "Unhandled error, shutting down: " 
	   << e.what() << endl;
    abort = true;
  } catch (...) {
    logger << "Unhandled error, shutting down." << endl; 
    abort = true;
  }

  close(intercomr[0]);
  close(intercomw[1]);

  // catch the dead baby
  if (waitpid(childspid, &result, 0) != childspid) {
    logger << "Authentication broke for <" << username 
	 << ">, waitpid returned unexpected value" << endl;
    string tmp = strerror(errno);
    return 1;
  }

  if (WIFSIGNALED(result)) {
    logger << "Authenticator died by signal "
	   << WTERMSIG(result) << " (";
    switch (WTERMSIG(result)) {
    case 1: logger << "SIGHUP"; break;
    case 2: logger << "SIGINT"; break;
    case 3: logger << "SIGQUIT"; break;
    case 4: logger << "SIGILL"; break;
    case 5: logger << "SIGTRAP"; break;
    case 6: logger << "SIGABRT"; break;
    case 7: logger << "SIGBUS"; break;
    case 8: logger << "SIGFPE"; break;
    case 9: logger << "SIGKILL"; break;      
    case 10: logger << "SIGUSR1"; break;      
    case 11: logger << "SIGSEGV"; break;      
    case 12: logger << "SIGUSR2"; break;      
    case 13: logger << "SIGPIPE"; break;      
    case 14: logger << "SIGALRM"; break;      
    case 15: logger << "SIGTERM"; break;      
    default: logger << "see man signal(7)"; break;
    }

    logger << ") for <" << username << ">" << endl;
    sleep(atoi(authpenalty));
    return 1;
  }

  if (abort)
    return 1;

  switch (WEXITSTATUS(result)) {
  case 0:  break;
  case 1:
    // authentication failed - sleep
    logger << "Authentication failed for <" << username 
	   << ">, wrong userid or password" << endl;
    sleep(atoi(authpenalty));
    return 2;
  case 2:
    // internal error
    logger << "Authentication broke for <" << username
	 << ">, authentication module returned 2 (internal error)" << endl;

    return 1;
  case 111:
  default:
    // internal error
    logger << "Authentication broke for <" << username
	 << ">, authentication module returned 111 (internal error)" << endl;

    return 1;
  }
 
  return 0;
}
