/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/io/io.h
 *  
 *  Description:
 *    Declaration of the IO class.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifndef io_h_included
#define io_h_included
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stack>
#include <string>
#include <iostream>
#include <iomanip>
#include <map>

#include <errno.h>
#include <stdio.h>
#include <syslog.h>

#include <ctype.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "../exceptions.h"
#include "../util/convert.h"

// #define DEBUG

namespace Binc {

  //----------------------------------------------------------------------
  class IO {

  protected:

    //--
    BincStream outputBuffer;

    //--
    std::stack<char> inputBuffer;
    int inputsize;
    bool inputlimit;

    //--
    std::string logprefix;

    //--
    int buffersize;
    int transfertimeout;

    //--
    int seqnr;
    int pid;
    bool flushesOnEndl;
    int mode;
    bool useLogPrefix;
    bool enabled;

    //--
    FILE *fpout;

    //--
    int select(int maxfd,
	       fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
	       int &timeout);
  public:

    //--
    enum {MODE_PLAIN = 0x01, MODE_SYSLOG = 0x02};

    //--
    void clear(void);
    void noFlushOnEndl(void);
    void flushOnEndl(void);
    void flushContent(void);
    void disable(void);
    void enable(void);
    void enableLogPrefix(void);
    void resetInput(void);
    void disableInputLimit(void);
    void enableInputLimit(void);

    //--
    inline void setBufferSize(int s)
    {
      buffersize = s;
    }

    //--
    inline void setTransferTimeout(int s)
    {
      transfertimeout = s;
    }

    //--
    inline void setLogPrefix(const std::string s_in)
    {
      logprefix = s_in;
    }

    //--
    template <class T> IO &operator << (const T &o)
    {
      using namespace ::std;

      if (useLogPrefix && outputBuffer.getSize() == 0) {
	outputBuffer << setw(5) << setfill('0') << pid;
	outputBuffer << " ";
	outputBuffer << setw(5) << setfill('0') << seqnr++;
	if (logprefix != "")
	  outputBuffer << " [" << logprefix << "]";
	outputBuffer << " ";
      }

      outputBuffer << o;

      if ((int) outputBuffer.getSize() >= buffersize)
	flushContent();

      return *this;
    }

    //--
    IO &operator << (std::ostream &(*man)(std::ostream &));

    //--
    void setModeSyslog(const std::string &servicename, int facility);

    //--
    void setFd(FILE *fpout);

    //--
    virtual void writeStr(const std::string s);

    //--
    virtual int readChar(int timeout = -1, bool retry = true);

    //--
    virtual int pending(void) const;

    //--
    void unReadChar(int c_in);

    //--
    void unReadChar(const std::string &s_in);

    //--
    IO();
    IO(FILE *fp);
    virtual ~IO();
  };

  //----------------------------------------------------------------------
  class IOFactory {

    //--
    std::map<int, IO *> ioers;

    //--
    IOFactory(void);

  public:

    //--
    static IOFactory &getInstance(void);

    //--
    void assign(int, IO *);

    //--
    IO &get(int);

    //--
    ~IOFactory(void);
  };
}

#endif
