/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/mailbox/mailbox.cc
 *  
 *  Description:
 *    Implementation of the Mailbox class.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>

#include "message.h"
#include "../io/io.h"
#include "mailbox.h"

using namespace ::std;
using namespace Binc;

//------------------------------------------------------------------------
PendingUpdates::PendingUpdates(void) : expunges(), flagupdates()
{
  recent = 0;
  exists = 0;

  newrecent = false;
  newexists = false;
}

//------------------------------------------------------------------------
PendingUpdates::~PendingUpdates(void)
{
}

//------------------------------------------------------------------------
void PendingUpdates::addExpunged(unsigned int uid)
{
  expunges.push_back(uid);
}

//------------------------------------------------------------------------
void PendingUpdates::addFlagUpdates(unsigned int uid, unsigned int flags)
{
  flagupdates[uid] = flags;
}

//------------------------------------------------------------------------
void PendingUpdates::setExists(unsigned int n)
{
  exists = n;
  newexists = true;
}

//------------------------------------------------------------------------
void PendingUpdates::setRecent(unsigned int n)
{
  recent = n;
  newrecent = true;
}

//------------------------------------------------------------------------
unsigned int PendingUpdates::getExists(void) const
{
  return exists;
}

//------------------------------------------------------------------------
unsigned int PendingUpdates::getRecent(void) const
{
  return recent;
}

//------------------------------------------------------------------------
bool PendingUpdates::newExists(void) const
{
  return newexists;
}

//------------------------------------------------------------------------
bool PendingUpdates::newRecent(void) const
{
  return newrecent;
}

//------------------------------------------------------------------------
PendingUpdates::expunged_const_iterator::expunged_const_iterator(void)
{
}

//------------------------------------------------------------------------
PendingUpdates::expunged_const_iterator::expunged_const_iterator(vector<unsigned int>::iterator i) : internal(i)
{
}

//------------------------------------------------------------------------
unsigned int PendingUpdates::expunged_const_iterator::operator * (void) const
{
  return *internal;
}

//------------------------------------------------------------------------
void PendingUpdates::expunged_const_iterator::operator ++ (void)
{
  ++internal;
}

//------------------------------------------------------------------------
bool PendingUpdates::expunged_const_iterator::operator == (expunged_const_iterator i) const
{
  return internal == i.internal;
}

//------------------------------------------------------------------------
bool PendingUpdates::expunged_const_iterator::operator != (expunged_const_iterator i) const
{
  return internal != i.internal;
}

//------------------------------------------------------------------------
PendingUpdates::expunged_const_iterator PendingUpdates::beginExpunged(void)
{
  return expunged_const_iterator(expunges.begin());
}

//------------------------------------------------------------------------
PendingUpdates::expunged_const_iterator PendingUpdates::endExpunged(void)
{
  return expunged_const_iterator(expunges.end());
}

//------------------------------------------------------------------------
PendingUpdates::flagupdates_const_iterator::flagupdates_const_iterator(void)
{
}

//------------------------------------------------------------------------
PendingUpdates::flagupdates_const_iterator::flagupdates_const_iterator(map<unsigned int, unsigned int>::iterator i) : internal(i)
{
}

//------------------------------------------------------------------------
void PendingUpdates::flagupdates_const_iterator::operator ++ (void)
{
  ++internal;
}

//------------------------------------------------------------------------
bool PendingUpdates::flagupdates_const_iterator::operator != (flagupdates_const_iterator i) const
{
  return internal != i.internal;
}

//------------------------------------------------------------------------
PendingUpdates::flagupdates_const_iterator PendingUpdates::beginFlagUpdates(void)
{
  return flagupdates_const_iterator(flagupdates.begin());
}

//------------------------------------------------------------------------
PendingUpdates::flagupdates_const_iterator PendingUpdates::endFlagUpdates(void)
{
  return flagupdates_const_iterator(flagupdates.end());
}

//------------------------------------------------------------------------
unsigned int PendingUpdates::flagupdates_const_iterator::first(void) const
{
  return internal->first;
}

//------------------------------------------------------------------------
unsigned int PendingUpdates::flagupdates_const_iterator::second(void) const
{
  return internal->second;
}

//------------------------------------------------------------------------
File::File(Message & m, FILE *f) : fp(f), message(m)
{
  ref = new unsigned int;
  *ref = 1;
}

//------------------------------------------------------------------------
File::File(const File &f) : fp(f.fp), message(f.message)
{
  ref = f.ref;
  ++(*ref);
}

//------------------------------------------------------------------------
File::operator FILE * () const
{
  return fp;
}


//------------------------------------------------------------------------
File::File & File::operator = (const File &f)
{
  if (&f != this) { // Handle self-assignment
    delete ref;
    fp = f.fp;
    ref = f.ref;
    ++(*ref);
  }

  return *this;
}
    
//------------------------------------------------------------------------
File::~File(void)
{
  --(*ref);
  if (*ref == 0 && fp != NULL) {
    delete ref;
    message.close();
    fclose(fp); // EK: check for errors
  }
  //  else
  //    logger << "~File invoked, but FILE* is still open." << fp << endl;
}

//------------------------------------------------------------------------
Mailbox::iterator::iterator(const Mailbox & d,
				 std::map<std::string, Message>::iterator i,
				 const BincImapParserSet &bs,
				 int mod)
  : mailbox(d),
    internalIterator(i),
    bset(bs),
    modifier(mod),
    sqnr(1),
    last(-1),
    currentErased(false)
{
  if (internalIterator != mailbox.messages.end())
    last = modifier & Mailbox::SQNR_MODE ?
      mailbox.getMaxSqnr() : mailbox.getMaxUid();
    
  reposition();
}

//------------------------------------------------------------------------
void Mailbox::iterator::reposition(void)
{
  if (modifier & Mailbox::SKIP_EXPUNGED) {
    while (internalIterator != mailbox.messages.end()
	   && internalIterator->second.isExpunged())
      ++internalIterator;
  }

  while (internalIterator != mailbox.messages.end()) {
    int num = modifier & Mailbox::SQNR_MODE ?
      sqnr : internalIterator->second.getUID();

    if (bset.isInSet(num)) break;
    if (num == last && !bset.isLimited()) break;
    if (!internalIterator->second.isExpunged())
      ++sqnr;
    ++internalIterator;
  }
}

//------------------------------------------------------------------------
Message & Mailbox::iterator::erase(void)
{
  std::map<std::string, Message>::iterator i = internalIterator;
  ++i;
  Message & message = internalIterator->second;
  mailbox.erase(message);
  internalIterator = i;
  currentErased = true;
  return message;
}

//------------------------------------------------------------------------
Message &Mailbox::iterator::operator *(void)
{
  if (currentErased) throw MessageException("Current message has been erased.");
  return internalIterator->second;
}
      
//------------------------------------------------------------------------
bool Mailbox::iterator::operator == (const iterator &msg) const
{
  return internalIterator == msg.internalIterator;
}

//------------------------------------------------------------------------
bool Mailbox::iterator::operator != (const iterator &msg) const
{
  return internalIterator != msg.internalIterator;
}
      
//------------------------------------------------------------------------
Mailbox::iterator & Mailbox::iterator::operator ++ (void)
{
  if (currentErased) {
    currentErased = false;
    return *this;
  }
  ++internalIterator;
  if (internalIterator != mailbox.messages.end() && 
      !internalIterator->second.isExpunged())
    ++sqnr;
  reposition();
  return *this;
}

//------------------------------------------------------------------------
Mailbox::iterator::~iterator(void)
{
};


//------------------------------------------------------------------------
Mailbox::Mailbox(void) : messages()
{
  readonly = false;
  cacheSize = 3;
  all =  BincImapParserSet();
  all.addRange(1, (unsigned int)-1);
}

//------------------------------------------------------------------------
Mailbox::~Mailbox(void)
{
  flushCache();
}

//------------------------------------------------------------------------
bool Mailbox::has(const std::string &id) const
{
  return messages.find(id) != messages.end();
}

//------------------------------------------------------------------------
Message & Mailbox::get(const std::string &id) const
{
  return messages.find(id)->second;
}

//------------------------------------------------------------------------
void Mailbox::add(const std::string &key, Message & m) const
{
  messages[key] = m;
}

//------------------------------------------------------------------------
void Mailbox::erase(Message & m) const
{
  messages.erase(m.getID());
}

//------------------------------------------------------------------------
Mailbox::iterator 
Mailbox::begin(const BincImapParserSet &bset,
		    int mod) const
{
  return Mailbox::iterator(*this, messages.begin(), bset, mod);
}

//------------------------------------------------------------------------
Mailbox::iterator Mailbox::end(void) const
{
  return Mailbox::iterator(*this, messages.end(), all, 0);
}

//------------------------------------------------------------------------
const File & Mailbox::getFileForMessage(Message &message)
{
  const int id = message.getUID();
  if (cached.find(id) == cached.end())
    addToCache(message);
  return *(cached.find(id)->second);
}

//------------------------------------------------------------------------
void Mailbox::setCacheSize(unsigned int size)
{
  cacheSize = size;
}

//------------------------------------------------------------------------
void Mailbox::flushCache(void)
{
  while (!cachedQ.empty()) {
    delete cached.find(cachedQ.front())->second;
    cached.erase(cachedQ.front());
    cachedQ.pop();
  }
}

//------------------------------------------------------------------------
void Mailbox::addToCache(Message &message) const
{
  int id = message.getUID();
  FILE * fp = fopen(message.getID()); 
  if (fp == NULL)
    throw MessageException("error opening file");    
  if (cachedQ.size() >= cacheSize) {
    delete cached.find(cachedQ.front())->second;
    cached.erase(cachedQ.front());
    cachedQ.pop();
  }
  cached.insert(make_pair(id, new File(message, fp)));
  cachedQ.push(id);
}

//------------------------------------------------------------------------
void Mailbox::parseFull(Message &message) const
{
  if (!message.isAllParsed()) {
    const int id = message.getUID();
    if (cached.find(id) == cached.end())
      addToCache(message);
    const File & f = *(cached.find(id)->second);
    fseek(f, 0, SEEK_SET);
    message.parseFull(f);
  }
}

//------------------------------------------------------------------------
void Mailbox::parseOnlyHeader(Message &message) const
{
  if (!message.isHeaderParsed() && !message.isAllParsed()) {
    const int id = message.getUID();
    if (cached.find(id) == cached.end())
      addToCache(message);
    const File & f = *(cached.find(id)->second);
    fseek(f, 0, SEEK_SET);
    message.parseOnlyHeader(f);
  }
}

//------------------------------------------------------------------------
int Mailbox::getMaxUid(void) const
{
  return getUidNext() - 1;
}
    
//------------------------------------------------------------------------
int Mailbox::getMaxSqnr(void) const
{
  map<string, Message>::const_iterator i = messages.begin();
  int sqnr = 0;
  while (i != messages.end()) {
    if (!i->second.isExpunged()) ++sqnr;
    ++i;
  }
  return sqnr;
}

//------------------------------------------------------------------------
void Mailbox::imapClose(void)
{
  flushCache();
  messages.clear();
}

Binc::BincImapParserSet Mailbox::all;
