/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-fetch.cc
 *  
 *  Description:
 *    Implementation of the FETCH command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>

#include "depot.h"
#include "io/io.h"
#include "mailbox/mailbox.h"
#include "operators.h"
#include "parsers/imap/imapparser.h"
#include "parsers/imap/recursivedescent/recursivedescent.h"
#include "session.h"
#include "util/convert.h"

using namespace ::std;
using namespace Binc;

namespace {
  //----------------------------------------------------------------------
  void printOneHeader(IO &io, const MimePart *message, const string &s_in,
			     bool removecomments = true)
  {
    string tmp = "";
    HeaderItem hitem;

    if (message->h.getFirstHeader(s_in, hitem)) {
      tmp = hitem.getValue();
      io << toImapString(unfold(tmp, removecomments));
    } else
      io << "NIL";
  }

  //----------------------------------------------------------------------
  void printOneAddressList(IO &io, const MimePart *message,
			   const string &s_in, bool removecomments = true)
  {
    string tmp = "";
    HeaderItem hitem;

    if (message->h.getFirstHeader(s_in, hitem)) {
      tmp = hitem.getValue();
      vector<string> addr;
      splitAddr(unfold(tmp, removecomments), addr);
      if (addr.size() != 0) {
	io << "(";
	for (vector<string>::const_iterator i = addr.begin();
	     i != addr.end(); ++i)
	  io << Address(*i).toParenList();
	io << ")";
      } else io << "NIL";
    } else
      io << "NIL";
  }

  //----------------------------------------------------------------------
  void envelope(IO &io, const MimePart *message)
  {
    HeaderItem hitem;
    io << "(";
    printOneHeader(io, message, "date");
    io << " ";
    printOneHeader(io, message, "subject", false);
    io << " ";
    printOneAddressList(io, message, "from", false);
    io << " ";
    printOneAddressList(io, message, message->h.getFirstHeader("sender", hitem) ? "sender" : "from", false);
    io << " ";
    printOneAddressList(io, message, message->h.getFirstHeader("reply-to", hitem) ? "reply-to" : "from", false);
    io << " ";
    printOneAddressList(io, message, "to", false);
    io << " ";
    printOneAddressList(io, message, "cc", false);
    io << " ";
    printOneAddressList(io, message, "bcc", false);
    io << " ";
    printOneHeader(io, message, "in-reply-to");
    io << " ";
    printOneHeader(io, message, "message-id");
    io << ")";
  }

  //----------------------------------------------------------------------
  void bodyStructure(IO &io, const MimePart *message, bool extended = true)
  {
    HeaderItem hitem;
    if (message->isMultipart() && message->members.size() > 0) {
      io << "(";
      
      for (vector<MimePart>::const_iterator i = message->members.begin();
	   i != message->members.end(); ++i)
	bodyStructure(io, &(*i));

      io << " ";
      io << toImapString(message->getSubType());
      io << " ";

      vector<string> parameters;
      vector<string> headers;
      string tmp;

      string type, subtype;

      tmp = "";
      if (message->h.getFirstHeader("content-type", hitem)) {
	tmp = unfold(hitem.getValue());
	trim(tmp);

	vector<string> v;
	split(tmp, ";", v);
	
	for (vector<string>::const_iterator i = v.begin(); i != v.end(); ++i) {
	  string element = *i;
	  trim(element);
	  if (element.find('=') != string::npos) {
	    string::size_type pos = element.find('=');
	    string s = element.substr(0, pos);
	    string t = element.substr(pos + 1);
	    trim(t, " \"");
	    parameters.push_back(s);
	    parameters.push_back(t);
	  }
	}

	if (parameters.size() != 0) {
	  io << "(";
	  for (vector<string>::const_iterator i = parameters.begin();
	       i != parameters.end(); ++i) {
	    if (i != parameters.begin())
	      io << " ";
	    io << toImapString(*i);
	  }
	  io << ")";
	} else
	  io << "NIL";
      } else
	io << "NIL";

      // CONTENT-DISPOSITION
      io << " ";
      tmp = "";
      if (message->h.getFirstHeader("content-disposition", hitem)) {
	tmp = hitem.getValue();
	trim(tmp);
	
	vector<string> v;
	split(tmp, ";", v);
	if (v.size() > 0) {
	  string disp = v[0];
	  trim(disp);
	  io << "(" << toImapString(disp);
	  io << " ";
	  if (v.size() > 1) {
	    io << "(";
	    vector<string>::const_iterator i = v.begin();
	    ++i;
	    bool wrote = false;
	    while (i != v.end()) {
	      string s = *i;
	      trim(s);
	      
	      string::size_type pos = s.find('=');
	      string key = s.substr(0, pos);
	      string value = s.substr(pos + 1);
	      trim(key);
	      trim(value);
	      
	      trim(key, " \"");
	      trim(value, " \"");
	      
	      if (!wrote) wrote = true;
	      else io << " ";
	      io << toImapString(key);
	      
	      io << " ";
	      io << toImapString(value);
	      
	      ++i;
	    }
	    io << ")";
	  } else 
	    io << "NIL";
	  io << ")";
	} else
	  io << "NIL";
      } else
	io << "NIL";
      
      // CONTENT-LANGUAGE
      io << " ";
      printOneHeader(io, message, "content-language");

      io << ")";
    } else {
      io << "(";

      vector<string> parameters;
      vector<string> headers;
      string tmp;
      tmp = "";
      string type, subtype;

      tmp = "";
      if (message->h.getFirstHeader("content-type", hitem)) {
	tmp = unfold(hitem.getValue());
	trim(tmp);
	
	vector<string> v;
	split(tmp, ";", v);

	if (v.size() > 0) {
	  vector<string> b;
	  split(v[0], "/", b);
	    
	  if (b.size() > 0)
	    type = b[0];
	  else
	    type = "text";

	  if (b.size() > 1)
	    subtype = b[1];
	  else
	    subtype = "plain";
	}
	
	for (vector<string>::const_iterator i = v.begin(); i != v.end(); ++i) {
	  if (i == v.begin()) continue;

	  string element = *i;
	  trim(element);

	  if (element.find('=') != string::npos) {
	    string::size_type pos = element.find('=');
	    string s = element.substr(0, pos);
	    string t = element.substr(pos + 1);
	    trim(t, " \"");
	    parameters.push_back(s);
	    parameters.push_back(t);
	  }
	}
      } else {
	type = "text";
	subtype = "plain";
      }

      io << toImapString(type);
      io << " ";
      io << toImapString(subtype);

      io << " ";
      if (parameters.size() != 0) {
	io << "(";
	for (vector<string>::const_iterator i = parameters.begin();
	     i != parameters.end(); ++i) {
	  if (i != parameters.begin())
	    io << " ";
	  io << toImapString(*i);
	}
	io << ")";
      } else
	io << "NIL";
      
      // CONTENT-ID
      io << " ";
      printOneHeader(io, message, "content-id");

      // CONTENT-DESCRIPTION
      io << " ";
      printOneHeader(io, message, "content-description");

      // CONTENT-TRANSFER-ENCODING
      io << " ";
      tmp = "";
      if (message->h.getFirstHeader("content-transfer-encoding", hitem)) {
	tmp = hitem.getValue();
	trim(tmp);
	io << toImapString(tmp);
      } else
	io << "\"7bit\"";
      io << " ";

      // Size of body in octets
      io << message->getBodyLength();

      lowercase(type);
      if (type == "text") {
	io << " ";
	io << message->getNofBodyLines();
      } else if (message->isMessageRFC822()) {
	io << " ";
	envelope(io, &message->members[0]);
	io << " ";
	bodyStructure(io, &message->members[0]);
	io << " ";
	io << message->getNofBodyLines();
      }
      
      // Extension data follows

      if (extended) {

	// CONTENT-MD5
	io << " ";
	printOneHeader(io, message, "content-md5");

	// CONTENT-DISPOSITION
	io << " ";
	tmp = "";
	if (message->h.getFirstHeader("content-disposition", hitem)) {
	  tmp = hitem.getValue();
	  trim(tmp);

	  vector<string> v;
	  split(tmp, ";", v);
	  if (v.size() > 0) {
	    string disp = v[0];
	    trim(disp);
	    io << "(" << toImapString(disp);
	    io << " ";
	    if (v.size() > 1) {
	      io << "(";
	      vector<string>::const_iterator i = v.begin();
	      ++i;
	      bool wrote = false;
	      while (i != v.end()) {
		string s = *i;
		trim(s);

		string::size_type pos = s.find('=');
		string key = s.substr(0, pos);
		string value = s.substr(pos + 1);
		trim(key);
		trim(value);

		trim(key, " \"");
		trim(value, " \"");

		if (!wrote) wrote = true;
		else io << " ";
		io << toImapString(key);

		io << " ";
		io << toImapString(value);

		++i;
	      }
	      io << ")";
	    } else 
	      io << "NIL";
	    io << ")";
	  } else
	    io << "NIL";
	} else
	  io << "NIL";
      
	// CONTENT-LANGUAGE
	io << " ";
	printOneHeader(io, message, "content-language");

	// CONTENT-LOCATION
	io << " ";
	printOneHeader(io, message, "content-location");

	io << ")";
      }
    }
  }

  void outputFlags(const Message & message) 
  {
    IO &com = IOFactory::getInstance().get(1);

    com << "FLAGS ";
	  
    com << "(";
    int flags = message.getStdFlags();
    vector<string> flagv;
    if (flags & Message::F_SEEN) flagv.push_back("\\Seen");
    if (flags & Message::F_ANSWERED) flagv.push_back("\\Answered");
    if (flags & Message::F_DELETED) flagv.push_back("\\Deleted");
    if (flags & Message::F_DRAFT) flagv.push_back("\\Draft");
    if (flags & Message::F_RECENT) flagv.push_back("\\Recent");
    if (flags & Message::F_FLAGGED) flagv.push_back("\\Flagged");
    for (vector<string>::const_iterator k 
	   = flagv.begin(); k != flagv.end(); ++k) {
      if (k != flagv.begin()) com << " ";
      com << *k;
    }
    com << ")";
  }

}

//----------------------------------------------------------------------
FetchOperator::FetchOperator(void)
{
}

//----------------------------------------------------------------------
FetchOperator::~FetchOperator(void)
{
}

//----------------------------------------------------------------------
const string FetchOperator::getName(void) const
{
  return "FETCH";
}

//----------------------------------------------------------------------
int FetchOperator::getState(void) const
{
  return Session::SELECTED;
}

//------------------------------------------------------------------------
void FetchOperator::process(Depot &depot, Command &command) const
{
  IO &com = IOFactory::getInstance().get(1);
  IO &logger = IOFactory::getInstance().get(2);
  Session &session = Session::getInstance();

  bool updateFlags = false;
  Command comm = command;

  Mailbox *mailbox = depot.getSelected();
  if (mailbox == NULL)
    throw CriticalException("fetch when no mailbox is selected");

  // If this is a UID FETCH, check if the UID attrobite is fetched. If
  // it is not, then add it to the list of fetch attributes.
  vector<BincImapParserFetchAtt>::const_iterator f_i;
  bool uidfetched = false;
  if (command.getUidMode()) {
    f_i = command.fatt.begin();
    while (f_i != command.fatt.end()) {
      if ((*f_i).type == "UID") {
	uidfetched = true;
	break;
      }
      
      f_i++;
    }

    if (!uidfetched) {
      BincImapParserFetchAtt b;
      b.type = "UID";
      comm.fatt.push_back(b);
    }
  }

  int mode;
  if (comm.getUidMode())
    mode = Mailbox::UID_MODE;
  else
    mode = Mailbox::SQNR_MODE;

  Mailbox::iterator i = mailbox->begin(comm.bset, 
					  Mailbox::SKIP_EXPUNGED | mode);
  Mailbox::iterator e = mailbox->end();

  for (;i != e;++i) {
    Message & message = *i;
    
    try {
      com << "* " << i.getSqnr() << " FETCH (";
      
      bool hasprinted = false;
      f_i = comm.fatt.begin();
      while (f_i != comm.fatt.end()) {
	BincImapParserFetchAtt fatt = *f_i;

	string prefix = "";
	if (hasprinted)
	  prefix = " ";
	
	if (fatt.type == "FLAGS") {
	  // FLAGS
	  hasprinted = true;
	  com << prefix;

	  outputFlags(message);
	} else if (fatt.type == "UID") {
	  // UID
	  com << prefix;
	  hasprinted = true;

	  com << "UID ";	  
	  com << message.getUID();
	} else if (fatt.type == "RFC822.SIZE") {
	  // RFC822.SIZE
	  com << prefix;
	  hasprinted = true;
	  if (message.getSize() == -1)
	    mailbox->parseFull(message);

	  com << "RFC822.SIZE ";
	  com << message.getSize();
	} else if (fatt.type == "ENVELOPE") {
	  // ENVELOPE
	  com << prefix;
	  hasprinted = true;
	  mailbox->parseOnlyHeader(message);	    

	  com << "ENVELOPE ";
	  envelope(com, &message);
	} else if (fatt.type == "BODYSTRUCTURE") {
	  // BODYSTRUCTURE
	  hasprinted = true;
	  com << prefix;
	  mailbox->parseFull(message);

	  com << "BODYSTRUCTURE ";
	  bodyStructure(com, &message);
	} else if (fatt.type == "BODY" && !fatt.hassection) {
	  // BODY with no section
	  hasprinted = true;
	  com << prefix;
	  mailbox->parseFull(message);
	  session.addBody();

	  com << "BODY ";
	  bodyStructure(com, &message, false);
	} else if (fatt.type == "INTERNALDATE") {
	  // INTERNALDATE
	  hasprinted = true;
	  com << prefix;

	  com << "INTERNALDATE ";
	  com << "\"" << message.getInternalDate() << "\"";
	} else if (fatt.type == "BODY" || fatt.type == "BODY.PEEK") {
	  // BODY & BODY.PEEK
	  com << prefix;
	  hasprinted = true;
	  session.addBody();

	  bool peek = (fatt.type == "BODY.PEEK");

	  com << fatt.toString();

	  bool includeheaders = true;
	  bool fullheader = false;
	  bool bodyfetch = false;

	  if (fatt.section != "" || fatt.sectiontext == "" || fatt.sectiontext == "TEXT") {
	    bodyfetch = true;
	    fullheader = true;
	  }

	  if (fatt.sectiontext == "HEADER.FIELDS.NOT")
	    includeheaders = false;

	  if (bodyfetch || fullheader)
	    mailbox->parseFull(message);
	  else
	    mailbox->parseOnlyHeader(message);

	  const MimePart *mptr = &message;
	  string genpart;

	  if (fatt.section == "")
	    mptr = (MimePart *) &message;
	  else
	    mptr = message.getPart(fatt.section, genpart);

	  if (mptr == NULL) {
	    // special case, returns empty content if invalid
	    // section was fetched.
	    com << "{0}\r\n";
	  } else {

	    const File &fp = mailbox->getFileForMessage(message);

	    if (fatt.sectiontext == "HEADER"
		|| fatt.sectiontext == "HEADER.FIELDS"
		|| fatt.sectiontext == "HEADER.FIELDS.NOT"
		|| fatt.sectiontext == "MIME") {
	      int size;
	      vector<string> v;
	      string dummy;

	      if (fatt.sectiontext == "MIME") {
		v.push_back("content-type");
		v.push_back("content-transfer-encoding");
		v.push_back("content-disposition");
		v.push_back("content-description");
	      }

	      size = fullheader 
		? mptr->printHeader(fp, com, v, true, 
				    fatt.offsetstart, 
				    fatt.offsetlength,
				    false, false, dummy)
		: mptr->printHeader(fp, com, fatt.headerlist, includeheaders, 
				    fatt.offsetstart,
				    fatt.offsetlength,
				    false, false, dummy);
	      
	      com << "{" << size << "}\r\n";
	      
	      size = fullheader 
		? mptr->printHeader(fp, com, v, true, 
				    fatt.offsetstart,
				    fatt.offsetlength,
				    true, false, dummy)
		: mptr->printHeader(fp, com, fatt.headerlist, includeheaders,
				    fatt.offsetstart, 
				    fatt.offsetlength,
				    true, false, dummy);
	    } else {
	      int size;
	      if (fatt.sectiontext == "" && fatt.section == "")
		size = mptr->printDoc(fp, com, fatt.offsetstart, 
				      fatt.offsetlength, false);
	      else
		size = mptr->printBody(fp, com, fatt.offsetstart, 
				       fatt.offsetlength, false);
		
	      com << "{" << size << "}\r\n";
		
	      if (fatt.sectiontext == "" && fatt.section == "")
		mptr->printDoc(fp, com, fatt.offsetstart, 
			       fatt.offsetlength, true);
	      else
		mptr->printBody(fp, com, fatt.offsetstart, 
				fatt.offsetlength, true);
	    }
	  }

	  // set the \Seen flag if .PEEK is not used.
	  if (!peek)
	    if ((message.getStdFlags() & Message::F_SEEN) == 0)
	      message.setStdFlag(Message::F_SEEN);

	} else if (fatt.type == "RFC822") {
	  com << prefix;
	  hasprinted = true;
	  session.addBody();

	  mailbox->parseFull(message);

	  com << fatt.toString();

	  const MimePart *mptr = &message;
	  vector<string> v;
	  string dummy;

	  const File &fp = mailbox->getFileForMessage(message);

	  int size = mptr->printDoc(fp, com, fatt.offsetstart, 
				    fatt.offsetlength, false);
	  com << " {" << size << "}\r\n";
	  mptr->printDoc(fp, com, fatt.offsetstart,
			 fatt.offsetlength, true);
	    
	  // set the \Seen flag
	  if ((message.getStdFlags() & Message::F_SEEN) == 0)
	    message.setStdFlag(Message::F_SEEN);

	} else if (fatt.type == "RFC822.HEADER") {
	  com << prefix;
	  hasprinted = true;

	  mailbox->parseFull(message);

	  com << fatt.toString();

	  const MimePart *mptr = &message;
	  vector<string> v;
	  string dummy;

	  const File &fp = mailbox->getFileForMessage(message);

	  int size = mptr->printHeader(fp, com, v, true, fatt.offsetstart, 
				       fatt.offsetlength, false, 
				       false, dummy);
	  com << " {" << size << "}\r\n";
	  size = mptr->printHeader(fp, com, v, true, fatt.offsetstart, 
				   fatt.offsetlength, true, false, dummy);
	    
	} else if (fatt.type == "RFC822.TEXT") {
	  // RFC822.TEXT
	  com << prefix;
	  hasprinted = true;
	  session.addBody();

	  mailbox->parseFull(message);

	  com << fatt.toString();

	  bool bodyfetch = false;
	  bodyfetch = true;

	  const MimePart *mptr = &message;
	  const File &fp = mailbox->getFileForMessage(message);

	  int size;
	  if (fatt.sectiontext == "" && fatt.section == "")
	    size = mptr->printDoc(fp, com, fatt.offsetstart, 
				  fatt.offsetlength, false);
	  else
	    size = mptr->printBody(fp, com, fatt.offsetstart, 
				   fatt.offsetlength, false);
	    
	  com << "{" << size << "}\r\n";
	    
	  if (fatt.sectiontext == "" && fatt.section == "")
	    mptr->printDoc(fp, com, fatt.offsetstart, 
			   fatt.offsetlength, true);
	  else
	    mptr->printBody(fp, com, fatt.offsetstart, 
			    fatt.offsetlength, true);

	  // set the \Seen flag
	  if ((message.getStdFlags() & Message::F_SEEN) == 0)
	    message.setStdFlag(Message::F_SEEN);
	} else {
	  // Unrecognized fetch_att, this is stopped by the parser
	  // so we never get here.
	}

	f_i++;
      }
    } catch (MessageException &e) {
      logger << "parse error: " << e.getReason() << endl;
      com << "* " << i.getSqnr()
	  << " NO [PARSE] FETCH failed, unable to parse message." << endl;
      continue;
    }

    com << ")" << endl;

    if (message.hasFlagsChanged()) {
      updateFlags = true;
      com << "* " << i.getSqnr() << " FETCH (";
      outputFlags(message);
      com << ")" << endl;
      message.setFlagsUnchanged();
    }
  }
  
  if (updateFlags) mailbox->updateFlags();
}

//----------------------------------------------------------------------
bool FetchOperator::parse(Command &c_in) const
{
  IO &com = IOFactory::getInstance().get(1);

  try {
    bool uidmode = false;
    if (expectThisString("UID")) {
      uidmode = true;
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
    }

    if (!expectThisString("FETCH")) {
      if (uidmode)
	com.unReadChar("UID ");
      uidmode = false;
      return false;
    }

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    if (!expectSet(c_in.getSet()))
      throw SyntaxException("expected set");

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    BincImapParserFetchAtt f;

    if (expectThisString("ALL")) {
      f.type = "ALL";
      c_in.fatt.push_back(f);
    } else if (expectThisString("FULL")) {
      f.type = "FULL";
      c_in.fatt.push_back(f);
    } else if (expectThisString("FAST")) {
      f.type = "FAST";
      c_in.fatt.push_back(f);
    } else if (expectFetchAtt(f)) {
      c_in.fatt.push_back(f);
    } else if (!expectThisString("(")) {
      throw SyntaxException("expected ALL, FULL, FAST, fetch_att or (");
    } else {
      while (1) {
	if (!expectFetchAtt(f))
	  throw SyntaxException("expected fetch_att");
	
	c_in.fatt.push_back(f);
	
	if (!expectSPACE())
	  break;
      }

      if (!expectThisString(")"))
	throw SyntaxException("expected )");
    }
    if (!expectCRLF())
      throw SyntaxException("expected CRLF");
    if (uidmode) c_in.setUidMode();
    c_in.setName("FETCH");
    return true;
  } catch (InputException &e) {
    e.addReason("in expectFETCH");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectFETCH");
    throw e;
  }
  return false;
}

//----------------------------------------------------------------------
bool FetchOperator::expectSectionText(BincImapParserFetchAtt &f_in) const
{
  try {
    if (expectThisString("HEADER")) {
      f_in.sectiontext = "HEADER";
      
      if (expectThisString(".FIELDS")) {
	f_in.sectiontext += ".FIELDS";
	
	if (expectThisString(".NOT"))
	  f_in.sectiontext += ".NOT";

	if (!expectSPACE())
	  throw SyntaxException("expected SPACE");

	if (!expectHeaderList(f_in))
	  throw SyntaxException("expected header_list");
      }
    } else if (expectThisString("TEXT"))
      f_in.sectiontext = "TEXT";
    else
      return false;

    return true;

  } catch (InputException &e) {
    e.addReason("in expectSectionText");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectSectionText");
    throw e;
  }
}

//----------------------------------------------------------------------
bool FetchOperator::expectSection(BincImapParserFetchAtt &f_in) const
{
  try {
    if (!expectThisString("["))
      return false;
    
    if (!expectSectionText(f_in)) {
      int n;
      if (expectNZNumber(n)) {

	BincStream nstr;

	nstr << n;

	f_in.section = nstr.str();
	
	bool gotadotalready = false;
	while (1) {
	  if (!expectThisString("."))
	    break;
	  
	  if (!expectNZNumber(n)) {
	    gotadotalready = true;
	    break;
	  }
	  
	  f_in.section += ".";	
	  
	  BincStream nstr;

	  nstr << n;

	  f_in.section += nstr.str();	 
	}
	
	if (gotadotalready || expectThisString(".")) {
	  if (expectThisString("MIME")) {
	    f_in.sectiontext = "MIME";
	  } else if (!expectSectionText(f_in))
	    throw SyntaxException("expected MIME or section_text");
	}
      }
    }

    if (!expectThisString("]"))
      throw SyntaxException("expected ]");
      
    return true;

  } catch (InputException &e) {
    e.addReason("in expectSection");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectSection");
    throw e;
  }
}

//----------------------------------------------------------------------
bool FetchOperator::expectHeaderList(BincImapParserFetchAtt &f_in) const
{
  try {
    if (!expectThisString("("))
      return false;

    string header_fld_name;
    while (1) {
      if (!expectAstring(header_fld_name))
	throw SyntaxException("expected header_fld_name");

      f_in.headerlist.push_back(header_fld_name);

      if (expectSPACE())
	continue;
      else break;
    }

    if (!expectThisString(")"))
      throw SyntaxException("expected )");

    return true;

  } catch (InputException &e) {
    e.addReason("in expectHeaderList");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectHeaderList");
    throw e;
  }
}

//----------------------------------------------------------------------
bool FetchOperator::expectOffset(BincImapParserFetchAtt &f_in) const
{
  try {
    if (!expectThisString("<"))
      return false;

    int i;
    if (!expectNumber(i))
      throw SyntaxException("expected number");

    if (!expectThisString("."))
      throw SyntaxException("expected .");

    int j;
    if (!expectNZNumber(j))
      throw SyntaxException("expected nz_number");

    if (!expectThisString(">"))
      throw SyntaxException("expected >");

    f_in.offsetstart = i;
    f_in.offsetlength = j;

    return true;

  } catch (InputException &e) {
    e.addReason("in expectOffset");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in expectOffset");
    throw e;
  }
}

//----------------------------------------------------------------------
bool FetchOperator::expectFetchAtt(BincImapParserFetchAtt &f_in) const
{
  try {
    if (expectThisString("ENVELOPE")) f_in.type = "ENVELOPE";
    else if (expectThisString("FLAGS")) f_in.type = "FLAGS";
    else if (expectThisString("INTERNALDATE")) f_in.type = "INTERNALDATE";
    else if (expectThisString("UID")) f_in.type = "UID";
    else if (expectThisString("RFC822")) {
      f_in.type = "RFC822";
      if (expectThisString(".HEADER")) f_in.type += ".HEADER";
      else if (expectThisString(".SIZE")) f_in.type += ".SIZE";
      else if (expectThisString(".TEXT")) f_in.type += ".TEXT";
      else if (expectThisString("."))
	throw SyntaxException("expected RFC822, RFC822.HEADER,"\
			      " RFC822.SIZE or RFC822.TEXT");

    } else if (expectThisString("BODY")) {
      f_in.type = "BODY";

      if (expectThisString("STRUCTURE")) f_in.type += "STRUCTURE";
      else if (expectThisString(".PEEK")) f_in.type += ".PEEK";
      
      if (!expectSection(f_in))
	f_in.hassection = false;
      else {
	f_in.hassection = true;
	expectOffset(f_in);
      } 
    } else
      return false;

    return true;
  } catch (InputException &e) {
    e.addReason("in FetchOperator::parse");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in FetchOperator::parse");
    throw e;
  }

  return true;
}
