/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-list.cc
 *  
 *  Description:
 *    Implementation of the LIST command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dirent.h>
#include <sys/stat.h>

#include <string>
#include <iostream>

#include "io/io.h"

#include "mailbox/mailbox.h"

#include "util/regmatch.h"
#include "util/convert.h"

#include "parsers/imap/recursivedescent/recursivedescent.h"

#include "session.h"
#include "depot.h"
#include "operators.h"

using namespace ::std;
using namespace Binc;

namespace {
  const int DIR_SELECT      = 0x01;
  const int DIR_MARKED      = 0x02;
  const int DIR_NOINFERIORS = 0x04;
}

//----------------------------------------------------------------------
ListOperator::ListOperator(void)
{
}

//----------------------------------------------------------------------
ListOperator::~ListOperator(void)
{
}

//----------------------------------------------------------------------
const string ListOperator::getName(void) const
{
  return "LIST";
}

//----------------------------------------------------------------------
int ListOperator::getState(void) const
{
  return Session::AUTHENTICATED | Session::SELECTED;
}

//------------------------------------------------------------------------
void ListOperator::process(Depot &depot, Command &command) const
{
  IO &com = IOFactory::getInstance().get(1);
  const char delim = depot.getDelimiter();

  // special case: if the mailbox argument is empty, then give a
  // hard coded reply.
  string wildcard;
  if ((wildcard = command.getListMailbox()) == "") {
    com << "* LIST (\\Noselect) \"" << delim << "\" \"\"" << endl;
    return;
  }

  // remove leading or trailing delimiter in wildcard
  trim(wildcard, string(&delim, 1));

  // convert wildcard to regular expression
  const string &regex = toRegex(wildcard, depot.getDelimiter());

  // remove leading or trailing delimiter in reference
  string ref = command.getMailbox();
  trim(ref, string(&delim, 1));

  // a map from mailbox name to flags
  map<string, int> mailboxes;

  // read through all entries in depository.
  for (Depot::iterator i = depot.begin("."); i != depot.end(); ++i) {
    const string path = *i;
    const string mpath = depot.filenameToMailbox(path);
    Mailbox *m = NULL;

    // skip entries that are not identified as mailboxes
    if ((m = depot.get(mpath)) == NULL)
      continue;

    // convert file name to mailbox name. skip it if there is no
    // corresponding mailbox name.
    string tmp = toCanonMailbox(depot.filenameToMailbox(path));
    trim(tmp, string(&delim, 1));
    if (tmp == "") continue;
    else {
      // inherit flags that were already set for this mailbox.
      int flags = DIR_SELECT;
      if (m->isMarked(path)) flags |= DIR_MARKED;
      if (mailboxes.find(tmp) != mailboxes.end()) flags |= mailboxes[tmp];
      mailboxes[tmp] = flags;
    }

    // now add all superior mailboxes with no flags set if not
    // added already.
    string::size_type pos = tmp.rfind(delim);
    while (pos != string::npos) {
      tmp = tmp.substr(0, pos);
      trim(tmp, string(&delim, 1));

      if (mailboxes.find(tmp) == mailboxes.end())
	mailboxes[tmp] = 0;

      pos = tmp.rfind(delim);
    }
  }

  // find leaf nodes O(N^2)
  map<string, int>::iterator i;
  for (i = mailboxes.begin(); i != mailboxes.end(); ++i) {
    string mailbox = i->first;
    mailbox += delim;

    bool leaf = true;
    map<string, int>::const_iterator j;
    for (j = mailboxes.begin(); j != mailboxes.end(); ++j) {
      string::size_type pos = j->first.rfind(delim);
      if (pos == string::npos) continue;

      string base = j->first.substr(0, pos + 1);

      if (mailbox == base) {
	leaf = false;
	break;
      }
    }

    if (leaf)
      i->second |= DIR_NOINFERIORS;
  }
  
  // finally, print all mailbox entries with flags.  
  for (i = mailboxes.begin(); i != mailboxes.end(); ++i) {
    if (ref == "" || ((ref.length() <= i->first.length()) 
		      && (i->first.substr(0, ref.length()) == ref)))
      if (regexMatch(i->first, regex) == 0) {
	com << "* LIST (";
	string sep = "";
	
	int flags = i->second;
	bool noselect = false;
	if (!(flags & DIR_SELECT)) {
	  com << sep << "\\Noselect";
	  sep = " ";
	  noselect = true;
	}
	
	if (!noselect) {
	  if (flags & DIR_MARKED)
	    com << sep << "\\Marked";
	  else
	    com << sep << "\\UnMarked";
	  sep = " ";
	}

	if (flags & DIR_NOINFERIORS)
	  com << sep << "\\Noinferiors";

	com << ") \"" << depot.getDelimiter() << "\" "
	    << toImapString(i->first) << endl;
      }
  }
}

//----------------------------------------------------------------------
bool ListOperator::parse(Command &c_in) const
{
  try {
    if (!expectThisString("LIST"))
      return false;

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    string mailbox;
    if (!expectMailbox(mailbox))
      throw SyntaxException("expected mailbox");
    c_in.setMailbox(mailbox);

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    string listmailbox;
    if (!expectListMailbox(listmailbox))
      throw SyntaxException("expected list_mailbox");
    if (!expectCRLF())
      throw SyntaxException("expected CRLF");
    c_in.setListMailbox(listmailbox);

  } catch (InputException &e) {
    e.addReason("in ListOperator::parse");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in ListOperator::parse");
    throw e;
  }

  c_in.setName("LIST");
  return true;
}
