/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/parsers/mime/mime.h
 *  
 *  Description:
 *    Declaration of main mime parser components
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef mime_h_included
#define mime_h_included
#include <string>
#include <vector>
#include <map>
#include <stdio.h>
#include "../../io/io.h"

namespace Binc {

  //---------------------------------------------------------------------- 
  class HeaderItem {
    mutable std::string key;
    mutable std::string value;
  public:
    inline const std::string &getKey(void) const
    {
      return key;
    }

    inline const std::string &getValue(void) const
    {
      return value;
    }

    HeaderItem(void);
    HeaderItem(const std::string &key, const std::string &value);
  };

  //---------------------------------------------------------------------- 
  class Header {
    mutable std::vector<HeaderItem> content;
  public:

    bool getFirstHeader(const std::string &key, HeaderItem &dest) const;
    bool getAllHeaders(const std::string &key, 
		       std::vector<HeaderItem> &dest) const;

    void add(const std::string &name, const std::string &content);
    void print(void) const;

    void clear(void) const;

    Header(void);
    ~Header(void);
  };

  //----------------------------------------------------------------------
  class MimeDocument;
  class MimePart {
  protected:
    mutable bool multipart;
    mutable bool messagerfc822;
    mutable std::string subtype;
    mutable std::string boundary;

    mutable int headerstartoffset;
    mutable int headerlength;

    mutable int bodystartoffset;
    mutable int bodystartoffsetcrlf;
    mutable int bodylength;
    mutable int nlines;
    mutable int nbodylines;
    mutable int size;

  public:
    mutable Header h;

    mutable std::vector<MimePart> members;

    inline const std::string &getSubType(void) const
    {
      return subtype;
    }

    inline bool isMultipart(void) const
    {
      return multipart;
    }

    inline bool isMessageRFC822(void) const
    {
      return messagerfc822;
    }

    inline int getSize(void) const
    {
      return bodylength;
    }

    inline int getNofLines(void) const
    {
      return nlines;
    }

    inline int getNofBodyLines(void) const
    {
      return nbodylines;
    }

    inline int getBodyLength(void) const
    {
      return bodylength;
    }

    inline int getBodyStartOffset(void) const
    {
      return bodystartoffset;
    }

    int printBody(FILE *fp, Binc::IO &output, int startoffset, 
		  int length, bool doprint) const;

    int printHeader(FILE *fp, Binc::IO &output, 
		    std::vector<std::string> headers, bool includeheaders, 
		    int startoffset, int length, bool doprint,
		    bool dostore, std::string &storage) const;

    int printDoc(FILE *fp, Binc::IO &output, int startoffset, 
		 int length, bool doprint) const;

    void printStruct(std::string s = "") const;

    virtual void clear(void) const;

    const MimePart *getPart(const std::string &findpart,
			    std::string genpart) const;

    virtual int parseOnlyHeader(FILE *fp, const std::string &toboundary, 
				int &bytenr, int &crlfnr) const;

    virtual int parseFull(FILE *fp, const std::string &toboundary, 
			  int &bytenr, int &crlfnr) const;

    MimePart(void);
    virtual ~MimePart(void);
  };

  //----------------------------------------------------------------------
  class MimeDocument : public MimePart {
    mutable bool headerIsParsed;
    mutable bool allIsParsed;

  public:
    void parseOnlyHeader(FILE *fp) const;
    void parseFull(FILE *fp) const;
    
    inline bool isHeaderParsed(void)
    {
      return headerIsParsed;
    }

    inline bool isAllParsed(void)
    {
      return allIsParsed;
    }

    void clear(void) const;

    MimeDocument(void);
    ~MimeDocument(void);
  };

};

#endif
